#![no_std]
#![no_main]

use ch32_hal::{self as hal, bind_interrupts};
use core::borrow::{Borrow, BorrowMut};
use core::cell::{Cell, RefCell};
use core::panic::PanicInfo;
use critical_section::Mutex;
use embassy_executor::Spawner;
use embassy_futures::join::join;
use embassy_time::{Duration, Ticker, Timer};
use embassy_usb::class::cdc_acm::{CdcAcmClass, State};
use embassy_usb::driver::EndpointError;
use embassy_usb::Builder;
use hal::gpio::{AnyPin, Level, Output, Pin};
use hal::pac;
use hal::pac::gpio::regs::Outdr;
use hal::pac::gpio::{self, Gpio};
use hal::println;
use hal::time::Hertz;
use hal::timer::low_level::CountingMode;
use hal::timer::simple_pwm::{PwmPin, SimplePwm};
use hal::{peripherals, usbd};
use heapless::{binary_heap::Min, BinaryHeap};
use micromath::F32Ext;

bind_interrupts!(struct Irqs {
    USB_LP_CAN1_RX0 => usbd::InterruptHandler<peripherals::USBD>;
});

const FPS: usize = 120;
const EMPTY_FRAME: Frame = Frame::empty();
static FRAME_QUEUE: Mutex<RefCell<BinaryHeap<Frame, Min, 30>>> =
    Mutex::new(RefCell::new(BinaryHeap::new()));

/// Unfortunately embedded_hal doesn't support atomic writes to multiple bytes, which is required for this implementation, so we must take a platform register :(
struct MBI5043<SDI, DCLK, LE> {
    sdi: SDI,
    dclk: DCLK,
    le: LE,
    port: Gpio,
}

#[repr(u32)]
enum DataType {
    Data = 1,
    Global = 3,
    ReadConfig = 5,
    WriteConfig = 11,
    EnableWriteConfig = 15,
}

pub trait PinBit {
    fn bit(&self) -> u32;
}

impl<T: Pin> PinBit for T {
    fn bit(&self) -> u32 {
        1 << self.pin()
    }
}

impl<'a, SDI: Pin, DCLK: Pin, LE: Pin> MBI5043<SDI, DCLK, LE> {
    pub fn new(sdi: SDI, dclk: DCLK, le: LE) -> Self {
        assert_eq!(sdi.port(), dclk.port());
        assert_eq!(sdi.port(), le.port());

        let port = pac::GPIO(sdi.port() as usize);
        Self::set_port_mode(&sdi);
        Self::set_port_mode(&dclk);
        Self::set_port_mode(&le);
        Self {
            port,
            sdi,
            dclk,
            le,
        }
    }
    fn set_port_mode<T: Pin>(pin: &T) {
        if pin.pin() < 8 {
            let reg = pac::GPIO(pin.port() as usize).cfglr();
            // low bits
            reg.modify(|w| {
                w.set_mode(pin.pin() as usize, gpio::vals::Mode::OUTPUT_50MHZ);
                w.set_cnf(
                    pin.pin() as usize,
                    gpio::vals::Cnf::ANALOG_IN__PUSH_PULL_OUT,
                );
            });
        } else {
            // high bits
            let reg = pac::GPIO(pin.port() as usize).cfghr();
            reg.modify(|r| {
                r.set_mode(pin.pin() as usize, gpio::vals::Mode::OUTPUT_50MHZ);
                r.set_cnf(
                    pin.pin() as usize,
                    gpio::vals::Cnf::ANALOG_IN__PUSH_PULL_OUT,
                );
            });
        }
    }
    fn clk_data_le_mask(&self) -> u32 {
        self.dclk.bit() | self.sdi.bit() | self.le.bit()
    }
    fn set_all(&self) {
        self.port.outdr().modify(|p| {
            p.set_odr(self.dclk.pin() as usize, true);
            p.set_odr(self.sdi.pin() as usize, true);
            p.set_odr(self.le.pin() as usize, true)
        });
    }
    fn send_word<const LE_LEN: u8>(&self, val: &u16) {
        for i in (0..16).rev() {
            // clk low & new data
            self.port.outdr().modify(|w| {
                w.set_odr(self.dclk.pin() as usize, false);
                w.set_odr(self.sdi.pin() as usize, val & (1 << i) != 0);
                w.set_odr(self.le.pin() as usize, i < LE_LEN);
            });

            // rising clock
            self.port
                .outdr()
                .modify(|w| w.set_odr(self.dclk.pin() as usize, true));
        }
        // clear all bits
        self.port
            .outdr()
            .modify(|w| w.0 &= !self.clk_data_le_mask());
    }
}

pub trait MBI {
    fn send_pixels<T: Fn(u8) -> u16>(&self, pixels: &[u8], transform: T);
}

impl<SDI: Pin, DCLK: Pin, LE: Pin> MBI for MBI5043<SDI, DCLK, LE> {
    fn send_pixels<T: Fn(u8) -> u16>(&self, pixels: &[u8], transform: T) {
        for pixel in pixels {
            self.send_word::<1>(&transform(*pixel));
        }
        self.send_word::<3>(&0x0000);
    }
}

#[embassy_executor::task]
async fn blink(pin1: AnyPin, pin2: AnyPin) {
    let mut leds = [
        Output::new(pin1, Level::Low, Default::default()),
        Output::new(pin2, Level::Low, Default::default()),
    ];
    let mut ticker = Ticker::every(Duration::from_millis(250));
    // println!("blinker running");
    let mut i = 0;
    loop {
        for idx in 0..leds.len() {
            leds[idx].set_level(((1 << idx) & i != 0).into());
        }
        i += 1;
        ticker.next().await
    }
}

// #[embassy_executor::task]
// async fn busy() {
//     loop {
//         pac::GPIOC.outdr().modify(|w| w.set_odr(13, !w.odr(13)));
//     }
// }

#[embassy_executor::task]
async fn usb(usb: peripherals::USBD, irq: Irqs, dp: peripherals::PA12, dm: peripherals::PA11) {
    let usb_driver = usbd::Driver::new(usb, irq, dp, dm);
    let usb_config = embassy_usb::Config::new(0xc0de, 0xcafe);

    let mut config_desc = [0; 256];
    let mut bos_desc = [0; 256];
    let mut control_buf = [0; 7];

    let mut state = State::new();
    let mut builder = Builder::new(
        usb_driver,
        usb_config,
        &mut config_desc,
        &mut bos_desc,
        &mut [],
        &mut control_buf,
    );

    let mut class = CdcAcmClass::new(&mut builder, &mut state, 64);
    let mut usb = builder.build();

    let usb_fut = usb.run();

    let echo_fut = async {
        loop {
            class.wait_connection().await;
            receive(&mut class).await;
        }
    };
    join(usb_fut, echo_fut).await;
}

struct Disconnected {}

impl From<EndpointError> for Disconnected {
    fn from(val: EndpointError) -> Self {
        match val {
            EndpointError::BufferOverflow => panic!("Buffer overflow"),
            EndpointError::Disabled => Disconnected {},
        }
    }
}

async fn receive<'d, T: usbd::Instance + 'd>(
    class: &mut CdcAcmClass<'d, usbd::Driver<'d, T>>,
) -> Result<(), Disconnected> {
    let mut buf = [0; 64];
    loop {
        let n = class.read_packet(&mut buf).await?;
        let data = &buf[..n];
        class.write_packet(data).await?;
        let mut frame = Frame::empty();
        frame.when = 512;
        let mut i = 0;
        for b in data {
            for j in 0..8 {
                if b & (1 << j) != 0 {
                    frame.pixels[i * 8 + j] = 128
                };
            }
            i += 1
        }
        println!("queueing new frame");
        critical_section::with(|cs| FRAME_QUEUE.borrow_ref_mut(cs).push(frame));
    }
}

#[derive(Ord, Eq)]
struct Frame {
    when: u64,
    pixels: [u8; 256],
}

impl Frame {
    const fn empty() -> Self {
        Self {
            when: 0,
            pixels: [0; 256],
        }
    }
}

impl Default for Frame {
    fn default() -> Self {
        Self::empty()
    }
}

impl PartialEq for Frame {
    fn eq(&self, other: &Self) -> bool {
        self.when.eq(&other.when)
    }
}

impl PartialOrd for Frame {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        self.when.partial_cmp(&other.when)
    }
}

struct GammaLut {
    gamma: f32,
    max: u16,
    lut: [u16; 256],
}

impl GammaLut {
    fn new(gamma: f32, max: u16) -> Self {
        let mut lut = [0u16; 256];
        let f32_max = max as f32;
        for i in 0..256 {
            let f_val = (i as f32 / 255.).powf(gamma) * f32_max + 0.5;
            lut[i] = f_val as u16;
        }
        Self { gamma, max, lut }
    }
    fn correct(&self, val: u8) -> u16 {
        self.lut[val as usize]
    }
}

struct Lightbar {
    gamma_lut: GammaLut,
    mbi: MBI5043<AnyPin, AnyPin, AnyPin>,
}

impl Lightbar {
    fn new(sdi: AnyPin, sck: AnyPin, le: AnyPin, gamma: f32) -> Self {
        Self {
            mbi: MBI5043::new(sdi, sck, le),
            gamma_lut: GammaLut::new(gamma, u16::MAX),
        }
    }
    async fn mbi_send_frame(&self, f: &Frame) {
        for row in 0..16 {
            pac::GPIOB.outdr().write_value(Outdr(0));
            self.mbi
                .send_pixels(&f.pixels[row * 16..(row + 1) * 16], |val| {
                    self.gamma_lut.correct(val)
                });
            pac::GPIOB.outdr().write_value(Outdr(1 << row));
            Timer::after_micros(1_000_000 / FPS as u64 / 17).await;
        }
    }
    async fn display_task(&self) {
        // configure GPIOB as output
        pac::GPIOB
            .cfglr()
            .write_value(gpio::regs::Cfglr(0x33333333));
        pac::GPIOB
            .cfghr()
            .write_value(gpio::regs::Cfghr(0x33333333));

        let mut frame: u64 = 0;
        let mut draw_frame: Frame = EMPTY_FRAME;

        let mut ticker = Ticker::every(Duration::from_micros(1_000_000 / FPS as u64));
        loop {
            critical_section::with(|cs| {
                let mut frames = FRAME_QUEUE.borrow_ref_mut(cs);
                if frames.peek().is_some_and(|f| frame >= f.when) {
                    draw_frame = frames.pop().unwrap();
                    println!("got new frame for {} @ {}", draw_frame.when, frame);
                }
            });
            self.mbi_send_frame(&draw_frame).await;
            frame += 1;
            ticker.next().await
        }
    }
}

#[embassy_executor::task]
async fn display_task(app: &'static Lightbar) {
    app.display_task().await
}

#[embassy_executor::main(entry = "qingke_rt::entry")]
// #[qingke_rt::entry]
async fn main(spawner: Spawner) -> ! {
    hal::debug::SDIPrint::enable();
    let mut config = hal::Config::default();
    config.rcc = hal::rcc::Config::SYSCLK_FREQ_144MHZ_HSE;

    let p = hal::init(config);
    Output::new(p.PC13.degrade(), Level::High, Default::default());

    let spi_sck = p.PA5;
    let spi_mosi = p.PA7;
    let le = p.PA4;
    let gclk = p.PA8;

    let ch1 = PwmPin::new_ch1::<0>(gclk);
    let mut pwm = SimplePwm::new(
        p.TIM1,
        Some(ch1),
        None,
        None,
        None,
        Hertz::mhz(8),
        CountingMode::default(),
    );
    let ch = hal::timer::Channel::Ch1;

    let max_duty = pwm.get_max_duty();
    pwm.set_duty(ch, 9);
    pwm.enable(ch);

    // println!("exti: {:x}", pac::EXTI.intenr().read().0);
    // pac::EXTI.intenr().modify(|w| w.set_mr(18, true));
    // println!("exti: {:x}", pac::EXTI.intenr().read().0);

    let app: Lightbar = Lightbar::new(spi_mosi.degrade(), spi_sck.degrade(), le.degrade(), 2.8);

    spawner.spawn(usb(p.USBD, Irqs, p.PA12, p.PA11)).unwrap();
    spawner
        .spawn(blink(p.PC14.degrade(), p.PC15.degrade()))
        .unwrap();

    // spawner
    //     .spawn(mbi_writer(
    //         spi_mosi.degrade(),
    //         spi_sck.degrade(),
    //         le.degrade(),
    //     ))
    //     .unwrap();

    // spawner.spawn(busy()).unwrap();

    critical_section::with(|cs| {
        FRAME_QUEUE.borrow_ref_mut(cs).push(Frame {
            when: 60,
            pixels: [128; 256],
        })
    });

    loop {
        app.display_task().await
    }
}

#[inline(never)]
#[panic_handler]
fn panic(info: &PanicInfo) -> ! {
    loop {
        core::sync::atomic::compiler_fence(core::sync::atomic::Ordering::SeqCst);
    }
}
